/*!
 * © 2021 David Vines
 *
 * Creative Commons License
 *
 * This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 */

/* globals ROT, Result, Move, State, Rikishi */
/* exported Opening, Contination, Bout */

const Opening = new State("Opening", self => {
	const cha = new Move("Charge",Rikishi.prototype.getSpeed,Rikishi.prototype.getWeight);
	cha.bothSameProhibited = function(result,us,them) {
		result.addText("Both "+us.toString()+" and "+them.toString()+" charge.");
		return false;
	};
	cha.addBothTriedMoveAndSucceededDescription = function(result) {
		result.addText("They crash heavily together in the middle of the dohyo.");
	};
	cha.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
		result.addText(us.toString()+" charges "+them.toString()+" but "+them.toString()+" absorbs the charge.");
	};
	cha.addWinDescription = function(result,us,them,_theirMove) {
		result.addText(us.toString()+" charges "+them.toString()+" pushing "+them.toString()+" all the way to the edge of the dohyo and out. "+us.toString()+" wins by Oshi-dashi.");
	};

	const att = new Move("Attack",Rikishi.prototype.getStrength,Rikishi.prototype.getStrength);
	att.bothSameProhibited = function(result,us,them) {
		result.addText("Both "+us.toString()+" and "+them.toString()+" go for the belt.");
		return false;
	};
	att.addBothTriedMoveAndSucceededDescription = function(result) {
		result.addText("They each get a grip on each other's belt.");
	};
	att.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
		result.addText(us.toString()+" grabs the belt of "+them.toString()+".");
	};
	att.addWinDescription = function(result,us,them,_theirMove) {
		result.addText(us.toString()+" grabs the belt of "+them.toString()+" and pushes "+them.toString()+" all the way to the edge of the dohyo and out of it. "+us.toString()+" wins by Yori-kiri.");
	};

	const ste = new Move("Step Aside",Rikishi.prototype.getAgility,Rikishi.prototype.getSpeed);
	ste.bothSameProhibited = function(_result,_us,_them) {
		return true;
	};
	ste.addBothTriedMoveAndSucceededDescription = function(_result) {
		throw new Error("Both rikishi succeeded in a step aside move - which the code should have prevented");
	};
	ste.addSuccessfulDescription = function(result,us,them,theirSuccessfulMove) {
		if (theirSuccessfulMove === Opening.Charge) {
			result.addText(us.toString()+" steps to the side of "+them.toString()+".");
		} else {
			result.addText(us.toString()+" steps to the side and behind "+them.toString()+".");
		}
	};
	ste.addWinDescription = function(result,us,them,_theirMove) {
		result.addText(us.toString()+" steps to the side and behind "+them.toString()+" pushing "+them.toString()+" out of the dohyo. "+us.toString()+" wins by Okuri-dashi.");
	};

	const sla = new Move("Slap",Rikishi.prototype.getStrength,Rikishi.prototype.getWeight);
	sla.bothSameProhibited = function(result,us,them) {
		result.addText("Both "+us.toString()+" and "+them.toString()+" come out slapping.");
		return false;
	};
	sla.addBothTriedMoveAndSucceededDescription = function(_result) {
	};
	sla.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
		result.addText(us.toString()+" slaps "+them.toString()+".");
	};
	sla.addWinDescription = function(result,us,them,theirMove) {
		if (theirMove === Opening.Charge) {
			result.addText(us.toString()+" slaps down "+them.toString()+" as they charge, winning by Hataki-komi.");
		} else {
			result.addText(us.toString()+" slaps down "+them.toString()+" winning by Tsuki-dashi.");
		}
	};

	cha.addCounterMove(cha, 85); cha.addCounterMove(att,-20); cha.addCounterMove(ste,460); cha.addCounterMove(sla, 10);
	att.addCounterMove(cha,460); att.addCounterMove(att, 85); att.addCounterMove(ste,-20); att.addCounterMove(sla, 60);
	ste.addCounterMove(cha,-20); ste.addCounterMove(att,460); ste.addCounterMove(ste, 85); ste.addCounterMove(sla,160);
	sla.addCounterMove(cha,160); sla.addCounterMove(att, 60); sla.addCounterMove(ste, 10); sla.addCounterMove(sla, 85);
	self.addCommonMoves([cha,att,ste,sla]);
});

const Continuation = new State("Opening", self => {
	const lev = new Move("Throw",Rikishi.prototype.getAgility,Rikishi.prototype.getStrength);
	lev.bothSameProhibited = function(_result,_us,_them) {
		return false;
	};
	lev.addBothTriedMoveAndSucceededDescription = function(result) {
		result.addText("Both rikishi attempt to throw their opponent, but neither gains the advantage.");
	};
	lev.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
		result.addText(us.toString()+" tries to throw "+them.toString()+".");
	};
	lev.addWinDescription = function(result,us,them,_theirMove) {
		result.addText(us.toString()+" throws "+them.toString()+". "+us.toString()+" wins by Uwate-nage.");
	};

	const lif = new Move("Lift",Rikishi.prototype.getStrength,Rikishi.prototype.getWeight);
	lif.bothSameProhibited = function(_result,_us,_them) {
		return false;
	};
	lif.addBothTriedMoveAndSucceededDescription = function(result) {
		result.addText("Both rikishi attempt to lift their opponent, but neither gains the advantage.");
	};
	lif.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
		result.addText(us.toString()+" attempts to lift "+them.toString()+".");
	};
	lif.addWinDescription = function(result,us,them,_theirMove) {
		result.addText(them.toString()+" is lifted out of the dohyo and "+us.toString()+" wins by Yori-taoshi.");
	};

	const pus = new Move("Push",Rikishi.prototype.getWeight,Rikishi.prototype.getWeight);
	pus.bothSameProhibited = function(_result,_us,_them) {
		return false;
	};
	pus.addBothTriedMoveAndSucceededDescription = function(result) {
		result.addText("Both rikishi attempt to push their opponent out of the dohyo, but neither gains the advantage.");
	};
	pus.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
		result.addText(us.toString()+" pushs "+them.toString()+", but "+them.toString()+" remains in the dohyo.");
	};
	pus.addWinDescription = function(result,us,them,_theirMove) {
		result.addText(them.toString()+" is pushed out of the dohyo and "+us.toString()+" wins by Oshi-dashi.");
	};

	const pul = new Move("Pull",Rikishi.prototype.getStrength,Rikishi.prototype.getAgility);
	pul.bothSameProhibited = function(_result,_us,_them) {
		return false;
	};
	pul.addBothTriedMoveAndSucceededDescription = function(result) {
		result.addText("Both rikishi attempt to pull their opponent to the ground, but neither gains the advantage.");
	};
	pul.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
		result.addText(us.toString()+" pulls on "+them.toString()+".");
	};
	pul.addWinDescription = function(result,us,them,_theirMove) {
		result.addText(them.toString()+" is pulled to the clay and "+us.toString()+" wins by Hiki-otoshi.");
	};

	const tri = new Move("Trip",Rikishi.prototype.getSpeed,Rikishi.prototype.getAgility);
	tri.bothSameProhibited = function(_result,_us,_them) {
		return false;
	};
	tri.addBothTriedMoveAndSucceededDescription = function(result) {
		result.addText("Both rikishi attempt to trip their opponent, but neither gains the advantage.");
	};
	tri.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
		result.addText(us.toString()+" tries to trip "+them.toString()+".");
	};
	tri.addWinDescription = function(result,us,them,_theirMove) {
		result.addText(us.toString()+" trips "+them.toString()+" and "+us.toString()+" wins by Soto-gake.");
	};


	lev.addCounterMove(lev, 85); lev.addCounterMove(lif,460); lev.addCounterMove(pus,-20); lev.addCounterMove(pul, 10); lev.addCounterMove(tri, 10);
	lif.addCounterMove(lev,-20); lif.addCounterMove(lif, 85); lif.addCounterMove(pus,460); lif.addCounterMove(pul, 60); lif.addCounterMove(tri,-20);
	pus.addCounterMove(lev,460); pus.addCounterMove(lif,-20); pus.addCounterMove(pus, 85); pus.addCounterMove(pul,160); pus.addCounterMove(tri,160);
	pul.addCounterMove(lev,160); pul.addCounterMove(lif, 60); pul.addCounterMove(pus, 10); pul.addCounterMove(pul, 85); pul.addCounterMove(tri, 85);
	tri.addCounterMove(lev,460); tri.addCounterMove(lif,-20); tri.addCounterMove(pus, 10); tri.addCounterMove(pul,460); tri.addCounterMove(tri, 85);
	self.addCommonMoves([lev,lif,pus,pul,tri]);
});

const BoutComplete = new State("BoutComplete", _self => {});

Opening.bothSucceeded = function(_eastmove,_westmove) {
	return Continuation;
};
Opening.succeeded = function(_rik,_move) {
	return Continuation;
};
Opening.neitherSucceeded = function(result) {
	result.addText("Both rikishi crash together from the tachi-ai and lock up in the middle of the dohyo with neither having the advantage.");
	return Continuation;
};
Continuation.bothSucceeded = function(_eastmove,_westmove) {
	return Continuation;
};
Continuation.succeeded = function(_rik,_move) {
	return Continuation;
};
Continuation.neitherSucceeded = function(result) {
	result.addText("Both rikishi remain locked up in the middle of the dohyo with neither having the advantage.");
	return Continuation;
};

class Bout {
	constructor(east,west,debug) {
		this._east = east;
		this._west = west;
		this._result = new Result();
		this._debug = debug;
	}
	resolve() {
		let time = 0;
		let state = Opening;
		while(state != BoutComplete) {
			state = this._resolveMoves(state,time++);
			this._result.newline();
		}
		return this._result;
	}
	_resolveMoves(state,time) {
		let eastmove, westmove, nextState;
		do {
			eastmove = this._east.getMove(this._west,state.getEastMoves(),time,this._result,this._debug);
			westmove = this._west.getMove(this._east,state.getWestMoves(),time,this._result,this._debug);
		}
		while(eastmove === westmove && eastmove.bothSameProhibited(this._result,this._east,this._west));
		if (this._debug) {
			this._result.newline();
			this._result.addText(`${state} ${this._east} ${eastmove} ${Math.floor(eastmove.getOdds(westmove,this._east,this._west))}%, ${this._west} ${westmove} ${Math.floor(westmove.getOdds(eastmove,this._west,this._east))}%`);
			this._result.addText(`${this._east} saves ${Math.floor(this._east.getSaveChance(time))}%, ${this._west} saves ${Math.floor(this._west.getSaveChance(time))}%`);
			this._result.newline();
		}
		const eastroll = ROT.RNG.getPercentage();
		const westroll = ROT.RNG.getPercentage();
		if (this._debug) {
			this._result.addText(`${this._east} rolls ${eastroll}, ${this._west} rolls ${westroll}.`);
		}
		const eastmovesuccess = (eastroll <= eastmove.getOdds(westmove,this._east,this._west));
		const westmovesuccess = (westroll <= westmove.getOdds(eastmove,this._west,this._east));
		const eastsaves = this._east.saves(time);
		const westsaves = this._west.saves(time);
		if (this._debug) {
			if (eastsaves) this._result.addText(` ${this._east} saves.`);
			if (westsaves) this._result.addText(` ${this._west} saves.`);
			this._result.newline();
		}
		const eastposswin = eastmovesuccess && !westsaves;
		const westposswin = westmovesuccess && !eastsaves;
		if (eastposswin && westposswin) {
			// Both rikishi succeeded and both failed to save - Need to determine who was the fastest
			const eastspeed = ROT.RNG.getPercentage()+this._east.getBalance()/2+ROT.RNG.getUniform()/5;
			const westspeed = ROT.RNG.getPercentage()+this._west.getBalance()/2+ROT.RNG.getUniform()/5;
			if (this._debug) {
				this._result.addText(`${this._east} speed is ${eastspeed}, ${this._west} speed is ${westspeed}.`);
				this._result.newline();
			}
			if (eastspeed > westspeed) {
				eastmove.addWinDescription(this._result,this._east,this._west,westmove);
				this._result.setWinner(this._east);
				this._result.setLoser(this._west);
				nextState = BoutComplete;
			} else {
				westmove.addWinDescription(this._result,this._west,this._east,eastmove);
				this._result.setWinner(this._west);
				this._result.setLoser(this._east);
				nextState = BoutComplete;
			}
		} else if (eastposswin) {
			eastmove.addWinDescription(this._result,this._east,this._west,this._westmove);
			this._result.setWinner(this._east);
			this._result.setLoser(this._west);
			nextState = BoutComplete;
		} else if (westposswin) {
			westmove.addWinDescription(this._result,this._west,this._east,eastmove);
			this._result.setWinner(this._west);
			this._result.setLoser(this._east);
			nextState = BoutComplete;
		} else if (eastmovesuccess && westmovesuccess) {
			// Both rikishi succeeded in their moves, but also both saved
			if (eastmove === westmove) {
				// Both made the same move
				eastmove.addBothTriedMoveAndSucceededDescription(this._result);
			} else {
				eastmove.addSuccessfulDescription(this._result,this._east,this._west,westmove);
				westmove.addSuccessfulDescription(this._result,this._west,this._east,eastmove);
			}
			nextState = state.bothSucceeded(eastmove,westmove);
		} else if (eastmovesuccess) {
			eastmove.addSuccessfulDescription(this._result,this._east,this._west,null);
			nextState = state.succeeded(this._east,eastmove);
		} else 	if (westmovesuccess) {
			westmove.addSuccessfulDescription(this._result,this._west,this._east,null);
			nextState = state.succeeded(this._west,westmove);
		} else {
			// Neither rikishi succeeded
			nextState = state.neitherSucceeded(this._result);
		}
		return nextState;
	}
}